//
// Copyright (c) Microsoft Corporation.  All rights reserved.
//
//
// Use of this source code is subject to the terms of the Microsoft end-user
// license agreement (EULA) under which you licensed this SOFTWARE PRODUCT.
// If you did not accept the terms of the EULA, you are not authorized to use
// this source code. For a copy of the EULA, please see the LICENSE.RTF on your
// install media.
//


// ****************************************************************************
// FILE: tinyclock.cpp
// ABTRACT: Main implementation file for the today item
// ****************************************************************************
//

#include "stdafx.h"

// definitions

#define ARRAYSIZEOF(x)			(sizeof(x) / sizeof((x)[0]))

#define PXSPERLINES				20		// number of pixels per today lines

#define DISPTYPE_TIME			0
#define DISPTYPE_DATE			1
#define DISPTYPE_TIMEANDDATE	2
#define DISPTYPE_DATEANDTIME	3
#define MAX_DISPTYPE			4

#define DISPPOS_LEFT			0
#define DISPPOS_CENTER			1
#define DISPPOS_RIGHT			2
#define MAX_DISPPOS				3

#define DISPORDER_TIME			0
#define DISPORDER_DATE			1
#define MAX_DISPORDER			2

#define FONTSEL_DEFAULT			0
#define FONTSEL_SAME			1
#define FONTSEL_DIFF			2
#define MAX_FONTSEL				3

// structures

typedef struct _DISPORDER {
	int count;
	int order[MAX_DISPORDER];
} DISPORDER, *PDISPORDER;

// constants

static const TCHAR *szDispType[MAX_DISPTYPE] = {
	TEXT("Time"),
	TEXT("Date"),
	TEXT("Time Date"),
	TEXT("Date Time"),
};
static const int	nDispType = ARRAYSIZEOF(szDispType);

static const DISPORDER aDispOrder[MAX_DISPTYPE] = {
	{ 1, { DISPORDER_TIME } },
	{ 1, { DISPORDER_DATE } },
	{ 2, { DISPORDER_TIME, DISPORDER_DATE } },
	{ 2, { DISPORDER_DATE, DISPORDER_TIME } },
};
static const int	nDispOrder = ARRAYSIZEOF(aDispOrder);

static const TCHAR *szDispPos[MAX_DISPPOS] = {
	TEXT("Left"),
	TEXT("Center"),
	TEXT("Right"),
};
static const int	nDispPos = ARRAYSIZEOF(szDispPos);

static const TCHAR *szFontSel[MAX_FONTSEL] = {
	TEXT("Use default font"),
	TEXT("Use same font for time/date"),
	TEXT("Use different font for time/date"),
};
static const int	nFontSel = ARRAYSIZEOF(szFontSel);

static const TCHAR *szTimeFormats[] = {
	TEXT("HH:mm"),
	TEXT("H:mm"),
	TEXT("hh:mm"),
	TEXT("h:mm"),
};
static const int	nTimeFormats = ARRAYSIZEOF(szTimeFormats);

static const TCHAR *szDateFormats[] = {
	TEXT("M/d"),
	TEXT("MM/dd"),
	TEXT("yy/M/d"),
	TEXT("yyyy/M/d"),
	TEXT("M/d/yy"),
	TEXT("M/d/yyyy"),
};
static const int	nDateFormats = ARRAYSIZEOF(szDateFormats);

static const TCHAR *szWeekFormats[] = {
	TEXT(""),
	TEXT("(w)"),
	TEXT("(YO)"),
};
static const int	nWeekFormats = ARRAYSIZEOF(szWeekFormats);

static const TCHAR *szDayOfTheWeek[] = {
	TEXT("Su"),
	TEXT("Mo"),
	TEXT("Tu"),
	TEXT("We"),
	TEXT("Th"),
	TEXT("Fr"),
	TEXT("Sa"),
};
static const int	nDayOfTheWeek = ARRAYSIZEOF(szDayOfTheWeek);

static const TCHAR *szYoubi[] = {
	TEXT(""),
	TEXT(""),
	TEXT(""),
	TEXT(""),
	TEXT(""),
	TEXT(""),
	TEXT("y"),
};
static const int	nYoubi = ARRAYSIZEOF(szYoubi);

// registry

#define REGISTRY_ROOT			TEXT("Software\\yasuho\\TinyClock")
#define VALUE_FONTSEL           TEXT("FontSel")
#define VALUE_TIMEFONT          TEXT("TimeFont")
#define VALUE_DATEFONT          TEXT("DateFont")
#define VALUE_USECOLOR          TEXT("UseColor")
#define VALUE_TEXTCOLOR         TEXT("TextColor")
#define VALUE_CUSTCOLORS        TEXT("CustColors")
#define VALUE_DISPTYPE			TEXT("DispType")
#define VALUE_DISPPOS			TEXT("DispPos")
#define VALUE_TIMEFORMAT		TEXT("TimeFormat")
#define VALUE_DATEFORMAT		TEXT("DateFormat")
#define VALUE_WEEKFORMAT		TEXT("WeekFormat")

// global variables    

HICON               g_hIcon;
HINSTANCE           g_hInst;
HWND                g_hWnd;
TCHAR				g_szAppName[MAX_PATH];

BOOL				g_bInitialized;
SYSTEMTIME			g_lasttime;		// time last updated
INT					g_nLines;		// # of lines in today item
BOOL				g_bFontUpdated;	// logfont has been updated so lines should be recalculated
INT					g_aveWidth;

BOOL				g_bUpdated;		// options are updated

// get/set from/to registry

INT					g_fontsel;
LOGFONT				g_timefont;
LOGFONT				g_datefont;
BOOL				g_usecolor;
COLORREF			g_textcolor;
COLORREF			g_custcolors[16];
INT					g_disptype;
INT					g_disppos;
TCHAR				g_timeformat[32];
TCHAR				g_dateformat[32];
TCHAR				g_weekformat[32];

// forward function declarations
static INT InitializeClasses();
LRESULT WINAPI CustomItemOptionsDlgProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam);

/************************************************************************/
/* Compute today lines from logfont                                     */
/************************************************************************/
static void ComputeTodayLines(HWND hwnd)
{
	HDC hDC;
	HFONT hFont;
	HFONT hFontOld;
	TEXTMETRIC tmTime;
	TEXTMETRIC tmDate;
	LONG maxHeight;

	if (!g_bFontUpdated)
		return;

	g_nLines = 1;
	hDC = GetDC(hwnd);

	// Retrieve height of time font
	hFont = CreateFontIndirect(&g_timefont);
    hFontOld = (HFONT)SelectObject(hDC, hFont);

	GetTextMetrics(hDC, &tmTime);

	SelectObject(hDC, hFontOld);
    DeleteObject(hFont);

	// Retrieve height of date font
	hFont = CreateFontIndirect(&g_datefont);
    hFontOld = (HFONT)SelectObject(hDC, hFont);

	GetTextMetrics(hDC, &tmDate);

	SelectObject(hDC, hFontOld);
    DeleteObject(hFont);

	// Compute today lines
	maxHeight = max(tmTime.tmHeight, tmDate.tmHeight);
	g_nLines = (maxHeight + DRA::SCALEY(PXSPERLINES) - 1) / DRA::SCALEY(PXSPERLINES);
	if (g_nLines <= 0)
	{
		g_nLines = 1;	// at least it should be greater than 0
	}

	// spacer
	g_aveWidth = min(tmTime.tmAveCharWidth, tmDate.tmAveCharWidth);

	ReleaseDC(hwnd, hDC);

	g_bFontUpdated = FALSE;
}

/************************************************************************/
/* Formatting time														*/
/************************************************************************/
static int FormatTime(TCHAR *pbuf, SYSTEMTIME *ptime)
{
	TCHAR *pbuf0 = pbuf;
	BOOL b12hr;
	TCHAR *pformat;

	b12hr = FALSE;

	pformat = g_timeformat;
	while (*pformat)
	{
		switch (*pformat)
		{
		default:
			*pbuf++ = *pformat++;
			break;
		case TEXT('H'):
			if (pformat[1] == TEXT('H'))
			{
				pbuf += wsprintf(pbuf, TEXT("%02d"), ptime->wHour);
				pformat += 2;
			}
			else
			{
				pbuf += wsprintf(pbuf, TEXT("%d"), ptime->wHour);
				pformat++;
			}
			break;
		case TEXT('h'):
			b12hr = TRUE;
			if (pformat[1] == TEXT('h'))
			{
				pbuf += wsprintf(pbuf, TEXT("%02d"), ptime->wHour % 12);
				pformat += 2;
			}
			else
			{
				pbuf += wsprintf(pbuf, TEXT("%d"), ptime->wHour % 12);
				pformat++;
			}
			break;
		case TEXT('m'):
			if (pformat[1] == TEXT('m'))
			{
				pbuf += wsprintf(pbuf, TEXT("%02d"), ptime->wMinute);
				pformat += 2;
			}
			else
			{
				*pbuf++ = *pformat++;
			}
			break;
		}
	}

	if (b12hr)
	{
		pbuf += wsprintf(pbuf, TEXT("%s"), ptime->wHour < 12 ? TEXT("AM") : TEXT("PM"));
	}

	*pbuf = TEXT('\0');

	return (int)(pbuf - pbuf0);
}

/************************************************************************/
/* Formatting date														*/
/************************************************************************/
static int FormatDate(TCHAR *pbuf, SYSTEMTIME *ptime)
{
	TCHAR *pbuf0 = pbuf;
	TCHAR *pformat;
	int index;

	// date
	pformat = g_dateformat;
	while (*pformat)
	{
		switch (*pformat)
		{
		default:
			*pbuf++ = *pformat++;
			break;
		case TEXT('y'):
			if (wcsncmp(&pformat[1], TEXT("yyy"), 3) == 0)
			{
				pbuf += wsprintf(pbuf, TEXT("%d"), ptime->wYear);
				pformat += 4;
			}
			else if (pformat[1] == TEXT('y'))
			{
				pbuf += wsprintf(pbuf, TEXT("%02d"), ptime->wYear % 100);
				pformat += 2;
			}
			else	// unknown
			{
				*pbuf++ = *pformat++;
			}
			break;
		case TEXT('M'):
			if (pformat[1] == TEXT('M'))
			{
				pbuf += wsprintf(pbuf, TEXT("%02d"), ptime->wMonth);
				pformat += 2;
			}
			else
			{
				pbuf += wsprintf(pbuf, TEXT("%d"), ptime->wMonth);
				pformat++;
			}
			break;
		case TEXT('d'):
			if (pformat[1] == TEXT('d'))
			{
				pbuf += wsprintf(pbuf, TEXT("%02d"), ptime->wDay);
				pformat += 2;
			}
			else
			{
				pbuf += wsprintf(pbuf, TEXT("%d"), ptime->wDay);
				pformat++;
			}
			break;
		}
	}

	// day of the week
	pformat = g_weekformat;
	while (*pformat)
	{
		switch (*pformat)
		{
		default:
			*pbuf++ = *pformat++;
			break;
		case TEXT('w'):
			index = ptime->wDayOfWeek;
			if (index < 0 || index >= nDayOfTheWeek) // for our safety
			{
				index = 0;
			}
			pbuf += wsprintf(pbuf, TEXT("%s"), szDayOfTheWeek[index]);
			pformat++;
			break;
		case TEXT('Y'):
			if (pformat[1] == TEXT('O'))
			{
				index = ptime->wDayOfWeek;
				if (index < 0 || index >= nYoubi) // for our safety
				{
					index = 0;
				}
				pbuf += wsprintf(pbuf, TEXT("%s"), szYoubi[index]);
				pformat += 2;
			}
			else
			{
				*pbuf++ = *pformat++;
			}
			break;
		}
	}

	*pbuf = TEXT('\0');

	return pbuf - pbuf0;
}

/************************************************************************/
/* Save settings to registry                                          */
/************************************************************************/
static BOOL SaveSettingsToRegistry()
{
    BOOL fret = FALSE;
    LONG lret;
    HKEY hkey = NULL;

    lret = RegCreateKeyEx(HKEY_CURRENT_USER, REGISTRY_ROOT, 0, NULL, 0, KEY_ALL_ACCESS, NULL, &hkey, NULL);
    if (lret == ERROR_SUCCESS)
    {
        lret = RegSetValueEx(hkey, VALUE_FONTSEL, 0, REG_DWORD, (LPBYTE)&g_fontsel, (DWORD)sizeof g_fontsel);
        if (lret != ERROR_SUCCESS)
        {
            goto out;
        }

        lret = RegSetValueEx(hkey, VALUE_TIMEFONT, 0, REG_BINARY, (LPBYTE)&g_timefont, (DWORD)sizeof g_timefont);
		if (lret != ERROR_SUCCESS)
        {
	        goto out;
		}

        lret = RegSetValueEx(hkey, VALUE_DATEFONT, 0, REG_BINARY, (LPBYTE)&g_datefont, (DWORD)sizeof g_datefont);
		if (lret != ERROR_SUCCESS)
        {
	        goto out;
		}

        lret = RegSetValueEx(hkey, VALUE_USECOLOR, 0, REG_DWORD, (LPBYTE)&g_usecolor, (DWORD)sizeof g_usecolor);
        if (lret != ERROR_SUCCESS)
        {
            goto out;
        }

        lret = RegSetValueEx(hkey, VALUE_TEXTCOLOR, 0, REG_DWORD, (LPBYTE)&g_textcolor, (DWORD)sizeof g_textcolor);
        if (lret != ERROR_SUCCESS)
        {
            goto out;
        }

        lret = RegSetValueEx(hkey, VALUE_CUSTCOLORS, 0, REG_BINARY, (LPBYTE)g_custcolors, (DWORD)sizeof g_custcolors);
        if (lret != ERROR_SUCCESS)
        {
            goto out;
        }

        lret = RegSetValueEx(hkey, VALUE_DISPTYPE, 0, REG_DWORD, (LPBYTE)&g_disptype, (DWORD)sizeof g_disptype);
        if (lret != ERROR_SUCCESS)
        {
            goto out;
        }

        lret = RegSetValueEx(hkey, VALUE_DISPPOS, 0, REG_DWORD, (LPBYTE)&g_disppos, (DWORD)sizeof g_disppos);
        if (lret != ERROR_SUCCESS)
        {
            goto out;
        }

        lret = RegSetValueEx(hkey, VALUE_TIMEFORMAT, 0, REG_SZ, (LPBYTE)&g_timeformat, (DWORD)sizeof g_timeformat);
        if (lret != ERROR_SUCCESS)
        {
            goto out;
        }

        lret = RegSetValueEx(hkey, VALUE_DATEFORMAT, 0, REG_SZ, (LPBYTE)&g_dateformat, (DWORD)sizeof g_dateformat);
        if (lret != ERROR_SUCCESS)
        {
            goto out;
        }

        lret = RegSetValueEx(hkey, VALUE_WEEKFORMAT, 0, REG_SZ, (LPBYTE)&g_weekformat, (DWORD)sizeof g_weekformat);
        if (lret != ERROR_SUCCESS)
        {
            goto out;
        }

		fret = TRUE;
    }

out:
	if (hkey) RegCloseKey(hkey);

    return fret;
}

/************************************************************************/
/* Load settings from registry                                          */
/************************************************************************/
static BOOL LoadSettingsFromRegistry()
{
    BOOL fret = FALSE;
    LONG lret;
    HKEY hkey = NULL;
    DWORD size;

    lret = RegOpenKeyEx(HKEY_CURRENT_USER, REGISTRY_ROOT, 0, KEY_READ, &hkey);
    if (lret == ERROR_SUCCESS)
    {
        size = sizeof(g_fontsel);
        lret = RegQueryValueEx(hkey, VALUE_FONTSEL, 0, NULL, (LPBYTE)&g_fontsel, &size);
        if (lret != ERROR_SUCCESS)
        {
            goto out;
        }

        size = sizeof(g_timefont);
	    lret = RegQueryValueEx(hkey, VALUE_TIMEFONT, 0, NULL, (LPBYTE)&g_timefont, &size);
		if (lret != ERROR_SUCCESS)
        {
	        goto out;
		}

        size = sizeof(g_datefont);
	    lret = RegQueryValueEx(hkey, VALUE_DATEFONT, 0, NULL, (LPBYTE)&g_datefont, &size);
		if (lret != ERROR_SUCCESS)
        {
	        goto out;
		}

		g_bFontUpdated = TRUE;

        size = sizeof(g_usecolor);
        lret = RegQueryValueEx(hkey, VALUE_USECOLOR, 0, NULL, (LPBYTE)&g_usecolor, &size);
        if (lret != ERROR_SUCCESS)
        {
            goto out;
        }

        size = sizeof(g_textcolor);
        lret = RegQueryValueEx(hkey, VALUE_TEXTCOLOR, 0, NULL, (LPBYTE)&g_textcolor, &size);
        if (lret != ERROR_SUCCESS)
        {
            goto out;
        }

        size = sizeof(g_custcolors);
        lret = RegQueryValueEx(hkey, VALUE_CUSTCOLORS, 0, NULL, (LPBYTE)g_custcolors, &size);
        if (lret != ERROR_SUCCESS)
        {
            goto out;
        }

        size = sizeof(g_disptype);
        lret = RegQueryValueEx(hkey, VALUE_DISPTYPE, 0, NULL, (LPBYTE)&g_disptype, &size);
        if (lret != ERROR_SUCCESS)
        {
            goto out;
        }

        size = sizeof(g_disppos);
        lret = RegQueryValueEx(hkey, VALUE_DISPPOS, 0, NULL, (LPBYTE)&g_disppos, &size);
        if (lret != ERROR_SUCCESS)
        {
            goto out;
        }

        size = sizeof(g_timeformat);
        lret = RegQueryValueEx(hkey, VALUE_TIMEFORMAT, 0, NULL, (LPBYTE)&g_timeformat, &size);
        if (lret != ERROR_SUCCESS)
        {
            goto out;
        }

        size = sizeof(g_dateformat);
        lret = RegQueryValueEx(hkey, VALUE_DATEFORMAT, 0, NULL, (LPBYTE)&g_dateformat, &size);
        if (lret != ERROR_SUCCESS)
        {
            goto out;
        }

        size = sizeof(g_weekformat);
        lret = RegQueryValueEx(hkey, VALUE_WEEKFORMAT, 0, NULL, (LPBYTE)&g_weekformat, &size);
        if (lret != ERROR_SUCCESS)
        {
            goto out;
        }

		fret = TRUE;
    }

out:
	if (hkey) RegCloseKey(hkey);

    return fret;
}

/************************************************************************/
/* Load default logfont													*/
/************************************************************************/
static void LoadDefaultLogfont(PLOGFONT plf)
{
	HFONT hSysFont = (HFONT) GetStockObject(SYSTEM_FONT);
	GetObject(hSysFont, sizeof(LOGFONT), plf);

	g_bFontUpdated = TRUE;
}

/************************************************************************/
/* Initialize font information											*/
/************************************************************************/
static void InitializeFontInfo()
{
	switch (g_fontsel)
	{
	case FONTSEL_DEFAULT:
		LoadDefaultLogfont(&g_timefont);
		LoadDefaultLogfont(&g_datefont);
		break;
	case FONTSEL_SAME:
		g_datefont = g_timefont;
		break;
	}
}

/************************************************************************/
/* Initialization														*/
/************************************************************************/
static void Initialize()
{
	if (g_bInitialized)
		return;

	g_fontsel = FONTSEL_DEFAULT;
	memset(&g_timefont, 0, sizeof g_timefont);
	memset(&g_datefont, 0, sizeof g_datefont);
	g_usecolor = FALSE;
	g_textcolor = 0;
	memset(g_custcolors, 0, sizeof g_custcolors);
	g_disptype = DISPTYPE_TIME;
	g_disppos = DISPPOS_LEFT;
	wcscpy(g_timeformat, szTimeFormats[0]);
	wcscpy(g_dateformat, szDateFormats[0]);
	wcscpy(g_weekformat, szWeekFormats[0]);

	g_bFontUpdated = TRUE;

	LoadSettingsFromRegistry();

	InitializeFontInfo();

	g_bInitialized = TRUE;
}

/************************************************************************/
/* DeInitialize															*/
/************************************************************************/
static void Deinitialize()
{
}

/*************************************************************************/
/* Entry point for the dll                                                 */
/*************************************************************************/
BOOL APIENTRY DllMain( HANDLE hModule, DWORD  ul_reason_for_call, LPVOID lpReserved)
{
    switch (ul_reason_for_call)
    {
    case DLL_PROCESS_ATTACH:            
        
        // The DLL is being loaded for the first time by a given process.
        // Perform per-process initialization here.  If the initialization
        // is successful, return TRUE; if unsuccessful, return FALSE.
        g_hInst = (HINSTANCE)hModule;
        
        //load the icon
        g_hIcon = (HICON)LoadImage(g_hInst, MAKEINTRESOURCE(IDI_DISPLAYICON), IMAGE_ICON, DRA::SCALEX(16), DRA::SCALEY(16) ,LR_DEFAULTCOLOR);
        
		//initilize the application class, and set the global window handle
		LoadString(g_hInst, IDS_TODAY_STORAGE_APPNAME, g_szAppName, sizeof(g_szAppName) / sizeof(TCHAR));
        UnregisterClass(g_szAppName, g_hInst);
        InitializeClasses();
        g_hWnd = 0;
        break;
        
    case DLL_PROCESS_DETACH:
        // The DLL is being unloaded by a given process.  Do any
        // per-process clean up here, such as undoing what was done in
        // DLL_PROCESS_ATTACH.    The return value is ignored.
        
        DestroyIcon(g_hIcon);
        
        UnregisterClass(g_szAppName, g_hInst);
        g_hInst = NULL;
        break;           
    }
    
    return TRUE;
}

/*************************************************************************/
/* Handle any messages that may be needed for the plugin                 */
/* Handled messages:                                                     */
/*        WM_TODAYCUSTOM_QUERYREFRESHCACHE                                 */
/*        WM_CREATE                                                         */
/*        WM_LBUTTONUP                                                     */
/*        WM_PAINT                                                         */
/*        WM_DESTROY                                                         */
/*************************************************************************/
LRESULT CALLBACK WndProc (HWND hwnd, UINT uimessage, WPARAM wParam, LPARAM lParam) 
{
	SYSTEMTIME curtime;

	switch (uimessage)
    {          
        //check to see if a refresh is required
    case WM_TODAYCUSTOM_QUERYREFRESHCACHE: 
        {
            TODAYLISTITEM *ptliItem;
            INT iItemHeight;
            
            BOOL    bReturn = FALSE;
            
            // get the pointer to the item from the Today screen
            ptliItem = (TODAYLISTITEM*)wParam;

            if ((NULL == ptliItem) || (WaitForSingleObject(SHELL_API_READY_EVENT, 0) == WAIT_TIMEOUT))
            {
				return FALSE;
            }

			// check updates
			GetLocalTime(&curtime);
			if (curtime.wHour != g_lasttime.wHour || curtime.wMinute != g_lasttime.wMinute)
			{
				bReturn = TRUE;
			}

            // adjust the height of the today item based on text height
			ComputeTodayLines(hwnd);
            iItemHeight = DRA::SCALEY(PXSPERLINES * g_nLines);
			if (iItemHeight <= 0)
			{
				iItemHeight = 1;	// for safety
			}

			if (0 == ptliItem->cyp)
            {
                ptliItem->cyp = iItemHeight;
                bReturn = TRUE;
            }
            
            return bReturn;
        }        
        
    case WM_CREATE:         
		Initialize();
        break;
        
    case WM_LBUTTONUP: 
		// DialogBox(g_hInst, MAKEINTRESOURCE(IDD_TODAY_CUSTOM), hwnd, (DLGPROC)CustomItemOptionsDlgProc);
        break;          

	case WM_PAINT: 
        PAINTSTRUCT     ps;
        RECT            rcWindBounds;
        RECT            rcMyBounds;
        RECT            rcTemp;
        HDC             hDC;
        HFONT           hFont;
        HFONT           hFontOld;
        COLORREF        crText;
		PDISPORDER      pdo;
		int             i;
        TCHAR           szTimeBuffer[256];
		int				leftmargin;
		int				width;

        GetWindowRect( hwnd, &rcWindBounds); 
        
        hDC = BeginPaint(hwnd, &ps);

		// create a custom rectangle relative to the client area
        rcMyBounds.left = 0;
        rcMyBounds.top = 0;
        //rcMyBounds.top = DRA::SCALEY(2);
		rcMyBounds.right = rcWindBounds.right - rcWindBounds.left;
        rcMyBounds.bottom = rcWindBounds.bottom - rcWindBounds.top;          
        
        // draw the icon on the screen
        SetBkMode(hDC, TRANSPARENT);
        DrawIcon(hDC, 2, 0, g_hIcon);

		// set text color
		if (g_usecolor)
		{
			crText = g_textcolor;
		}
		else
		{
		    // determine the theme's current text color
		    //  this color will change when the user picks a new theme,
			//  so get it each time the display needs to be painted
			crText = SendMessage(GetParent(hwnd), TODAYM_GETCOLOR, (WPARAM) TODAYCOLOR_TEXT, NULL);
		}

        // retrieve current time
		GetLocalTime(&curtime);
		g_lasttime = curtime;

		// calculate display position
		leftmargin = DRA::SCALEX(28);
	    rcMyBounds.left += leftmargin;

		if (g_disppos == DISPPOS_LEFT)
		{
			// nothing needed
		}
		else
		{	// center or right
			rcTemp = rcWindBounds;	// whole window
			pdo = (PDISPORDER)&aDispOrder[g_disptype];
			for (i = 0; i < pdo->count; i++)
			{
				switch (pdo->order[i])
				{
				case DISPORDER_TIME:
					hFont = CreateFontIndirect(&g_timefont);
					hFontOld = (HFONT) SelectObject(hDC, hFont);
					FormatTime(szTimeBuffer, &curtime);
			        DrawText(hDC, szTimeBuffer, -1, &rcTemp, DT_CALCRECT); // compute rectangle
					rcTemp.left = rcTemp.right;
			        SelectObject(hDC, hFontOld);
					DeleteObject(hFont);
					break;
				case DISPORDER_DATE:
					hFont = CreateFontIndirect(&g_datefont);
					hFontOld = (HFONT) SelectObject(hDC, hFont);
					FormatDate(szTimeBuffer, &curtime);
			        DrawText(hDC, szTimeBuffer, -1, &rcTemp, DT_CALCRECT); // compute rectangle
					rcTemp.left = rcTemp.right;
			        SelectObject(hDC, hFontOld);
					DeleteObject(hFont);
					break;
				}
			}
			if (pdo->count >= 2)
				rcTemp.left += g_aveWidth;

			// adjust start position
			width = rcMyBounds.right - rcTemp.left;
			if (g_disppos == DISPPOS_CENTER)
			{
				width /= 2;	// centering
				if (width > leftmargin)
					rcMyBounds.left += (width - leftmargin);
			}
			else // DISPPOS_RIGHT
			{
				width -= DRA::SCALEX(2);	// right margin
				if (width > leftmargin)
					rcMyBounds.left += (width - leftmargin);
			}
		}

		// display time/date
		pdo = (PDISPORDER)&aDispOrder[g_disptype];
		for (i = 0; i < pdo->count; i++)
		{
			switch (pdo->order[i])
			{
			case DISPORDER_TIME:
		        hFont = CreateFontIndirect(&g_timefont);
		        hFontOld = (HFONT) SelectObject(hDC, hFont);

				FormatTime(szTimeBuffer, &curtime);

				// draw text and update rect for next
				rcTemp = rcMyBounds;
		        SetTextColor(hDC, crText);
		        DrawText(hDC, szTimeBuffer, -1, &rcTemp, DT_LEFT | DT_VCENTER);
		        DrawText(hDC, szTimeBuffer, -1, &rcTemp, DT_CALCRECT); // compute rectangle
				rcMyBounds.left = rcTemp.right + g_aveWidth;

		        SelectObject(hDC, hFontOld);
				DeleteObject(hFont);
				break;
			case DISPORDER_DATE:
		        hFont = CreateFontIndirect(&g_datefont);
		        hFontOld = (HFONT) SelectObject(hDC, hFont);

				FormatDate(szTimeBuffer, &curtime);

				// draw text and update rect for next 
				rcTemp = rcMyBounds;
		        SetTextColor(hDC, crText);
		        DrawText(hDC, szTimeBuffer, -1, &rcTemp, DT_LEFT | DT_VCENTER);
		        DrawText(hDC, szTimeBuffer, -1, &rcTemp, DT_CALCRECT); // compute rectangle
				rcMyBounds.left = rcTemp.right + g_aveWidth;

		        SelectObject(hDC, hFontOld);
				DeleteObject(hFont);
				break;
			}
		}
        EndPaint(hwnd, &ps);
        return 0;

	case WM_SIZE:
		RECT rc;

		// some mobile device does not refresh the screen when screen was rotated
		GetWindowRect(hwnd, &rc);
		InvalidateRect(hwnd, &rc, TRUE);
		break;

    case WM_DESTROY :  
		Deinitialize();
        return 0 ;

    // this fills in the background with the today screen image
    case WM_ERASEBKGND:
        TODAYDRAWWATERMARKINFO dwi;
        dwi.hdc = (HDC)wParam;
        GetClientRect(hwnd, &dwi.rc);
        
        dwi.hwnd = hwnd;
        SendMessage(GetParent(hwnd), TODAYM_DRAWWATERMARK, 0,(LPARAM)&dwi);
        return TRUE;
    }

    return DefWindowProc (hwnd, uimessage, wParam, lParam) ;
}


/*************************************************************************/
/* Create and register our window class for the today item                 */
/*************************************************************************/
INT InitializeClasses()
{
    WNDCLASS         wc; 
    memset(&wc, 0, sizeof(wc));
    
    wc.style         = 0;                   
    wc.lpfnWndProc     = (WNDPROC) WndProc;
    wc.hInstance     = g_hInst;
    wc.hIcon         = 0;
    wc.hCursor         = 0;
    wc.hbrBackground = (HBRUSH) GetStockObject(WHITE_BRUSH);
    wc.lpszClassName = g_szAppName;
    
    //register our window
    if(!RegisterClass(&wc))
    { 
        return 0 ;
    }
    return 1;
}

/*************************************************************************/
/* Initialize the DLL by creating a new window                             */
/*************************************************************************/
HWND InitializeCustomItem(TODAYLISTITEM *ptli, HWND hwndParent) 
{
    LPCTSTR appName = g_szAppName;
    
    //create a new window
    g_hWnd = CreateWindow(appName,appName,WS_VISIBLE | WS_CHILD,
        CW_USEDEFAULT,CW_USEDEFAULT,240,0,hwndParent, NULL, g_hInst, NULL) ;
    
    // attach our winproc to the newly created window
    SetWindowLong(g_hWnd, GWL_WNDPROC, (LONG) WndProc);
    
    //display the window
    ShowWindow (g_hWnd, SW_SHOWNORMAL);
    UpdateWindow (g_hWnd) ;  
    
	return g_hWnd;
}

/*************************************************************************/
/* Update color button information                                       */
/*************************************************************************/
void UpdateColorButtonInfo(HWND hDlg)
{
	if (g_usecolor)
	{
		SendMessage(GetDlgItem(hDlg, IDC_USECOLOR), BM_SETCHECK, BST_CHECKED, 0);
		EnableWindow(GetDlgItem(hDlg, IDC_COLOR), TRUE);
	}
	else
	{
		SendMessage(GetDlgItem(hDlg, IDC_USECOLOR), BM_SETCHECK, BST_UNCHECKED, 0);
		EnableWindow(GetDlgItem(hDlg, IDC_COLOR), FALSE);
	}
}

/*************************************************************************/
/* Update display type information                                       */
/*************************************************************************/
void UpdateDispTypeInfo(HWND hDlg)
{
	SendMessage(GetDlgItem(hDlg, IDC_DISPTYPE), CB_SETCURSEL, g_disptype, 0);
}

/*************************************************************************/
/* Update display position information                                   */
/*************************************************************************/
void UpdateDispPosInfo(HWND hDlg)
{
	SendMessage(GetDlgItem(hDlg, IDC_DISPPOS), CB_SETCURSEL, g_disppos, 0);
}

/*************************************************************************/
/* Update display format position information                            */
/*************************************************************************/
void UpdateFormatInfo(HWND hDlg, TCHAR *pFormat, const TCHAR *szFormat[], int nFormat)
{
	int nPos = 0;

	for (int i = 0; i < nFormat; i++)
	{
		if (wcscmp(pFormat, szFormat[i]) == 0)
		{
			nPos = i;
			break;
		}
	}
	if (nPos >= nFormat)
		nPos = 0;

	SendMessage(hDlg, CB_SETCURSEL, nPos, 0);
}

/*************************************************************************/
/* Update display format button information                              */
/*************************************************************************/
void UpdateFormatButtonInfo(HWND hDlg)
{
	switch (g_disptype)
	{
	case DISPTYPE_TIME:
		EnableWindow(GetDlgItem(hDlg, IDC_TIMEFORMAT), TRUE);
		EnableWindow(GetDlgItem(hDlg, IDC_DATEFORMAT), FALSE);
		EnableWindow(GetDlgItem(hDlg, IDC_WEEKFORMAT), FALSE);
		break;
	case DISPTYPE_DATE:
		EnableWindow(GetDlgItem(hDlg, IDC_TIMEFORMAT), FALSE);
		EnableWindow(GetDlgItem(hDlg, IDC_DATEFORMAT), TRUE);
		EnableWindow(GetDlgItem(hDlg, IDC_WEEKFORMAT), TRUE);
		break;
	case DISPTYPE_TIMEANDDATE:
	case DISPTYPE_DATEANDTIME:
		EnableWindow(GetDlgItem(hDlg, IDC_TIMEFORMAT), TRUE);
		EnableWindow(GetDlgItem(hDlg, IDC_DATEFORMAT), TRUE);
		EnableWindow(GetDlgItem(hDlg, IDC_WEEKFORMAT), TRUE);
		break;
	}
}

/*************************************************************************/
/* Update font selection information                                     */
/*************************************************************************/
void UpdateFontSelInfo(HWND hDlg)
{
	SendMessage(GetDlgItem(hDlg, IDC_FONTSEL), CB_SETCURSEL, g_fontsel, 0);
	switch (g_fontsel)
	{
	case FONTSEL_DEFAULT:
		EnableWindow(GetDlgItem(hDlg, IDC_TIMEFONT), FALSE);
		EnableWindow(GetDlgItem(hDlg, IDC_DATEFONT), FALSE);
		break;
	case FONTSEL_SAME:
		EnableWindow(GetDlgItem(hDlg, IDC_TIMEFONT), TRUE);
		EnableWindow(GetDlgItem(hDlg, IDC_DATEFONT), FALSE);
		break;
	case FONTSEL_DIFF:
		EnableWindow(GetDlgItem(hDlg, IDC_TIMEFONT), TRUE);
		EnableWindow(GetDlgItem(hDlg, IDC_DATEFONT), TRUE);
		break;
	}
}

/*************************************************************************/
/* Display preview on the option dialog									 */
/*************************************************************************/
void DisplayPreview(HWND hDlg)
{
	SYSTEMTIME curtime;
	PDISPORDER pdo;
	TCHAR *pBuffer;
	int i;
	TCHAR szBuffer[256];

	//GetLocalTime(&curtime);
	curtime.wHour = 1;
	curtime.wMinute = 2;
	curtime.wMonth = 3;
	curtime.wDay = 4;
	curtime.wYear = 2005;
	curtime.wDayOfWeek = 0; // Sunday

	pBuffer = szBuffer;
	pdo = (PDISPORDER)&aDispOrder[g_disptype];
	for (i = 0; i < pdo->count; i++)
	{
		switch (pdo->order[i])
		{
		case DISPORDER_TIME:
			pBuffer += FormatTime(pBuffer, &curtime);
			*pBuffer++ = TEXT(' ');
			*pBuffer = TEXT('\0');
			break;
		case DISPORDER_DATE:
			pBuffer += FormatDate(pBuffer, &curtime);
			*pBuffer++ = TEXT(' ');
			*pBuffer = TEXT('\0');
			break;
		}
	}

	SendDlgItemMessage(hDlg, IDC_PREVIEW, WM_SETTEXT, 0, (LPARAM)szBuffer);
}

/*************************************************************************/
/* Message Handler for display format dialog                             */
/*************************************************************************/
LRESULT WINAPI DispFormatDlgProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	int i;

	switch (message)
    {
    case WM_INITDIALOG:
        {
            SHINITDLGINFO shidi;

            // Create a Done button and size it.  
            shidi.dwMask = SHIDIM_FLAGS;
            shidi.dwFlags = SHIDIF_DONEBUTTON | SHIDIF_SIPDOWN | SHIDIF_SIZEDLGFULLSCREEN;
            shidi.hDlg = hDlg;
            SHInitDialog(&shidi);

			// setup list items
			for (i = 0; i < nTimeFormats; i++)
			{
				SendDlgItemMessage(hDlg, IDC_TIMEFORMAT, CB_ADDSTRING, 0, (LPARAM)szTimeFormats[i]);
			}
			for (i = 0; i < nDateFormats; i++)
			{
				SendDlgItemMessage(hDlg, IDC_DATEFORMAT, CB_ADDSTRING, 0, (LPARAM)szDateFormats[i]);
			}
			for (i = 0; i < nWeekFormats; i++)
			{
				SendDlgItemMessage(hDlg, IDC_WEEKFORMAT, CB_ADDSTRING, 0, (LPARAM)szWeekFormats[i]);
			}

			// update info
			UpdateFormatInfo(GetDlgItem(hDlg, IDC_TIMEFORMAT), g_timeformat, szTimeFormats, nTimeFormats);
			UpdateFormatInfo(GetDlgItem(hDlg, IDC_DATEFORMAT), g_dateformat, szDateFormats, nDateFormats);
			UpdateFormatInfo(GetDlgItem(hDlg, IDC_WEEKFORMAT), g_weekformat, szWeekFormats, nWeekFormats);

			UpdateFormatButtonInfo(hDlg);

			DisplayPreview(hDlg);
		}
		return TRUE;

    case WM_COMMAND:
        switch (LOWORD(wParam))
        {
		case IDOK:
            EndDialog(hDlg, LOWORD(wParam));
            return TRUE;

		case IDC_TIMEFORMAT:
			switch (HIWORD(wParam))
			{
			case CBN_SELCHANGE:
				SendDlgItemMessage(hDlg, IDC_TIMEFORMAT, WM_GETTEXT, ARRAYSIZEOF(g_timeformat), (LPARAM)g_timeformat);
				DisplayPreview(hDlg);
				g_bUpdated = TRUE;
				break;
			}
			break;

		case IDC_DATEFORMAT:
			switch (HIWORD(wParam))
			{
			case CBN_SELCHANGE:
				SendDlgItemMessage(hDlg, IDC_DATEFORMAT, WM_GETTEXT, ARRAYSIZEOF(g_dateformat), (LPARAM)g_dateformat);
				DisplayPreview(hDlg);
				g_bUpdated = TRUE;
				break;
			}
			break;

		case IDC_WEEKFORMAT:
			switch (HIWORD(wParam))
			{
			case CBN_SELCHANGE:
				SendDlgItemMessage(hDlg, IDC_WEEKFORMAT, WM_GETTEXT, ARRAYSIZEOF(g_weekformat), (LPARAM)g_weekformat);
				DisplayPreview(hDlg);
				g_bUpdated = TRUE;
				break;
			}
			break;

		}
		break;

	case WM_DESTROY:
        break;
	}

    return FALSE;
}

/*************************************************************************/
/* Message Handler for display font dialog                               */
/*************************************************************************/
LRESULT WINAPI FontsDlgProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	HWND hFontSel;
	int i;
	CHOOSEFONT cf;
	LOGFONT lf;

    switch (message)
    {
    case WM_INITDIALOG:
        {
            SHINITDLGINFO shidi;

            // Create a Done button and size it.  
            shidi.dwMask = SHIDIM_FLAGS;
            shidi.dwFlags = SHIDIF_DONEBUTTON | SHIDIF_SIPDOWN | SHIDIF_SIZEDLGFULLSCREEN;
            shidi.hDlg = hDlg;
            SHInitDialog(&shidi);

			hFontSel = GetDlgItem(hDlg, IDC_FONTSEL);
			for (i = 0; i < nFontSel; i++)
			{
				SendMessage(hFontSel, CB_ADDSTRING, 0, (LPARAM)szFontSel[i]);
			}

			UpdateFontSelInfo(hDlg);
			UpdateColorButtonInfo(hDlg);
        }
        return TRUE; 
        
    case WM_COMMAND:
        switch (LOWORD(wParam))
        {
		case IDOK:
            EndDialog(hDlg, LOWORD(wParam));
            return TRUE;

		case IDC_FONTSEL:
			switch (HIWORD(wParam))
			{
			case CBN_SELCHANGE:
				hFontSel = GetDlgItem(hDlg, IDC_FONTSEL);
				i = SendMessage(hFontSel, CB_GETCURSEL, 0, 0);
				if (i < 0 || i >= MAX_FONTSEL)
					break;
				g_fontsel = i;

				InitializeFontInfo();
				g_bFontUpdated = TRUE;

				UpdateFontSelInfo(hDlg);

				g_bUpdated = TRUE;
				break;
			}
			break;

		case IDC_TIMEFONT:
			lf = g_timefont;
			memset(&cf, 0, sizeof cf);
			cf.lStructSize = sizeof cf;
			cf.hInstance = g_hInst;
			cf.hwndOwner = hDlg;
			cf.lpLogFont = &lf;
			cf.Flags = CF_SCREENFONTS | CF_INITTOLOGFONTSTRUCT;
			if (!ChooseFont(&cf))
				break;

			g_timefont = lf;
			InitializeFontInfo();
			g_bFontUpdated = TRUE;

			g_bUpdated = TRUE;
			break;

		case IDC_DATEFONT:
			lf = g_datefont;
			memset(&cf, 0, sizeof cf);
			cf.lStructSize = sizeof cf;
			cf.hInstance = g_hInst;
			cf.hwndOwner = hDlg;
			cf.lpLogFont = &lf;
			cf.Flags = CF_SCREENFONTS | CF_INITTOLOGFONTSTRUCT;
			if (!ChooseFont(&cf))
				break;

			g_datefont = lf;
			InitializeFontInfo();
			g_bFontUpdated = TRUE;

			g_bUpdated = TRUE;
			break;

		case IDC_USECOLOR:
			g_usecolor = !g_usecolor;
			UpdateColorButtonInfo(hDlg);
			g_bUpdated = TRUE;
			break;

		case IDC_COLOR:
			CHOOSECOLOR cc;

			memset(&cc, 0, sizeof cc);
			cc.lStructSize = sizeof cc;
			cc.hInstance = g_hInst;
			cc.hwndOwner = hDlg;
			cc.rgbResult = g_textcolor;
			cc.lpCustColors = g_custcolors;
			cc.Flags = CC_RGBINIT | CC_ANYCOLOR;
			if (!ChooseColor(&cc))
				break;

			g_textcolor = cc.rgbResult;
			g_bUpdated = TRUE;
			break;

        }
        break;

    case WM_DESTROY:
        break;

    }

    return FALSE;
}

/*************************************************************************/
/* Message Handler for the options dialog                                 */
/*************************************************************************/
LRESULT WINAPI CustomItemOptionsDlgProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	HWND hDispType;
	HWND hDispPos;
	int i;

    switch (message)
    {
    case WM_INITDIALOG:
        {
            SHINITDLGINFO shidi;

            // Create a Done button and size it.  
            shidi.dwMask = SHIDIM_FLAGS;
            shidi.dwFlags = SHIDIF_DONEBUTTON | SHIDIF_SIPDOWN | SHIDIF_SIZEDLGFULLSCREEN;
            shidi.hDlg = hDlg;
            SHInitDialog(&shidi);

			Initialize();

			hDispType = GetDlgItem(hDlg, IDC_DISPTYPE);
			for (i = 0; i < nDispType; i++)
			{
				SendMessage(hDispType, CB_ADDSTRING, 0, (LPARAM)szDispType[i]);
			}

			hDispPos = GetDlgItem(hDlg, IDC_DISPPOS);
			for (i = 0; i < nDispPos; i++)
			{
				SendMessage(hDispPos, CB_ADDSTRING, 0, (LPARAM)szDispPos[i]);
			}

			//UpdateFontButtonInfo(hDlg);
			UpdateColorButtonInfo(hDlg);
			UpdateDispTypeInfo(hDlg);
			UpdateDispPosInfo(hDlg);

			DisplayPreview(hDlg);

			g_bUpdated = FALSE;
        }
        return TRUE; 
        
    case WM_COMMAND:
        switch (LOWORD(wParam))
        {
		case IDOK:
			if (g_bUpdated)
				SaveSettingsToRegistry();

            EndDialog(hDlg, LOWORD(wParam));
            return TRUE;

		case IDC_DISPTYPE:
			switch (HIWORD(wParam))
			{
			case CBN_SELCHANGE:
				hDispType = GetDlgItem(hDlg, IDC_DISPTYPE);
				i = SendMessage(hDispType, CB_GETCURSEL, 0, 0);
				if (i < 0 || i >= MAX_DISPTYPE)
					break;
				g_disptype = i;

				DisplayPreview(hDlg);

				g_bUpdated = TRUE;
				break;
			}
			break;

		case IDC_DISPPOS:
			switch (HIWORD(wParam))
			{
			case CBN_SELCHANGE:
				hDispPos = GetDlgItem(hDlg, IDC_DISPPOS);
				i = SendMessage(hDispPos, CB_GETCURSEL, 0, 0);
				if (i < 0 || i >= MAX_DISPPOS)
					break;
				g_disppos = i;

				DisplayPreview(hDlg);

				g_bUpdated = TRUE;
				break;
			}
			break;

		case IDC_DISPFORMAT:
			DialogBox(g_hInst, MAKEINTRESOURCE(IDD_DISPFORMAT), hDlg, (DLGPROC)DispFormatDlgProc);
			DisplayPreview(hDlg);
			break;

		case IDC_FONTS:
			DialogBox(g_hInst, MAKEINTRESOURCE(IDD_FONTS), hDlg, (DLGPROC)FontsDlgProc);
			//DisplayPreview(hDlg);
			break;

        }
        break;

    case WM_DESTROY:
        break;

    }

    return FALSE;
}


// vi: ts=4 sw=4
